##
## Version number is as follows: CFmajor.CFminor.tclcfMajor.tclcfMinor
##  Note -- at this time CloudFlare only appears to be using major version
##  		numbers, so minor will be 0.
##
package provide tclcf 4.0.0.1

package require Tcl 8.6
package require http
package require tls
package require json
package require json::write
package require oo::util

::http::register https 443 ::tls::socket

::oo::class create tclCloudFlareProtocol {
	variable KeyDict

	method CheckResults {resultDict} {
		if {![dict exists $resultDict success] || ![dict get $resultDict success]} {
			set errorMessage {}
			set errorCodes {}
			if {[dict exists $resultDict errors]} {
				set errorCodes [dict get $resultDict errors]
			} else {
				set errorCodes UNKNOWN
			}
			if {[dict exists $resultDict messages]} {
				set errorMessage [dict get $resultDict messages]
			}
			return -code error -errorcode [list tclCloudFlare CFERROR $errorCodes] "CloudFlare error: '$errorMessage'"
		}
	}

	method MakeRequest {httpMethod url {dataPacket {}}} {
		variable KeyDict

		set headers [list \
						X-Auth-Email [dict get $KeyDict emailAddr] \
						X-Auth-Key [dict get $KeyDict key] \
		]
		try {
			switch -exact -- $httpMethod {
				DELETE {
					set token [::http::geturl $url -headers $headers -method POST -query $dataPacket -type application/json]
				}
				GET {
					set token [::http::geturl $url -headers $headers -method GET -type application/json]
				}
				default {
					return \
						-code error \
						-errorcode [list tclCloudFlareZone UNSUPMETHOD $httpMethod] \
						"Unsupported method '$httpMethod'"
				}
			}
			::http::wait $token
			switch -exact -- [::http::ncode $token] {
				200 {
					set resultsDict [::json::json2dict [::http::data $token]]
					my CheckResults $resultsDict
				}
				default {
					array set meta [::http::meta $token]
					puts stderr "\n\nclock [clock seconds] [clock format [clock seconds] -gmt yes] RayID $meta(CF-Ray)"
					puts stderr "getURL $url headers $headers query $dataPacket"
					set code [::http::code $token]
					return \
						-code error \
						-errorcode [list tclCloudFlareZone UNEXPCODE [::http::ncode $token] $code] \
						"Unexpected HTTP code: '$code'.  Body: {[::http::data $token]}"
				}
			}
		} finally {
			::http::cleanup $token
		}

		return $resultsDict
	}


}

::oo::class create tclCloudFlareZone {
	superclass tclCloudFlareProtocol

	variable ZoneKeyUrlFmt
	variable ZoneInfo
	
	constructor {email key zoneID} {
		variable KeyDict
		variable ZoneInfo
		variable ZoneKeyUrlFmt

		set ZoneKeyUrlFmt {
			purgeCache {https://api.cloudflare.com/client/v4/zones/%1$s/purge_cache}
			getInfo {https://api.cloudflare.com/client/v4/zones/%1$s}
			delete {"https://api.cloudflare.com/client/v4/zones/%1$s}
		}

		dict set KeyDict emailAddr $email
		dict set KeyDict key $key
		dict set KeyDict zoneID $zoneID

		set url [format [dict get $ZoneKeyUrlFmt getInfo] [dict get $KeyDict zoneID]]

		try {
			set ZoneInfo [dict get [my MakeRequest GET $url] result]
		} trap {tclCloudFlareZone UNEXPCODE} {} {
			return -code error -errorcode [list tclCloudFlare UNKZONE $zoneID] -level 1 "Unknown CloudFlare Zone '$zoneID'"
		}

		return;
	}

	method delete {} {
		variable KeyDict
		variable ZoneKeyUrlFmt
		
		set url [format [dict get $ZoneKeyUrlFmt delete] [dict get $KeyDict zoneID]]

		MakeRequest DELETE $url
		
		return;
	}

	method info {args} {
		variable ZoneInfo

		if {[llength $args]} {
			return [dict get $ZoneInfo {*}$args]
		} else {
			return $ZoneInfo
		}

	}

	method purge_everything {} {
		variable KeyDict
		variable ZoneKeyUrlFmt

		set data [json::write object \
				  purge_everything true \
		]

		set url [format [dict get $ZoneKeyUrlFmt purgeCache] [dict get $KeyDict zoneID]]

		my MakeRequest DELETE $url $data

	}

	method purge_files {purgeDict} {
		variable KeyDict
		variable ZoneKeyUrlFmt

		set fileLists {}
		set tagList {}
		set jsonPurgeDict {}
		set send no

		if {[dict exists $purgeDict files]} {
			foreach file [dict get $purgeDict files] {
				lappend fileLists [json::write string $file]
			}

			##
			## Remove duplicates
			##
			set fileLists [lsort -dictionary -unique $fileLists]
			if {[llength $fileLists]} {
				dict set jsonPurgeDict files [json::write array {*}$fileLists]
				set send yes
			}
		}

		if {[dict exists $purgeDict tags]} {
			foreach tag [dict get $purgeDict tags] {
				lappend tagLists [json::write string $tag]
			}

			##
			## Remove duplicates
			##
			set tagLists [lsort -dictionary -unique $tagList]
			if {[llength $tagLists]} {
				dict set jsonPurgeDict tags [json::write array {*}$tagLists]
				set send yes
			}
		}

		if {$send} {
			set data [json::write object {*}$jsonPurgeDict]
		}
		
		set url [format [dict get $ZoneKeyUrlFmt purgeCache] [dict get $KeyDict zoneID]]

		my MakeRequest DELETE $url $data

	}

}


::oo::class create tclCloudFlare {
	superclass tclCloudFlareProtocol
	
	variable Cache
		
	constructor {email key} {
		variable KeyDict
		variable Cache

		dict set KeyDict emailAddr $email
		dict set KeyDict key $key
		set Cache {}

		return
	}
	
	destructor {
		variable Cache
		
		foreach objType [dict keys $Cache] {
			dict for {id object} [dict get $Cache $objType] {
				$object destroy
			}
		}	
		
	}
	
	method GetZones {} {
		set results {}

		set urlFmt {https://api.cloudflare.com/client/v4/zones?page=%d}
		set pageNum 1
		set totalProcessed 0
		set totalCount 1
		
		while {$totalProcessed < $totalCount} {
			set resultDict [my MakeRequest GET [format $urlFmt $pageNum]]
			set thisResultList [dict get $resultDict result]
			set results [concat $results $thisResultList]
			incr pageNum
			if {$totalProcessed == 0} {
				set totalCount [dict get $resultDict result_info total_count]
			}
			incr totalProcessed [dict get $resultDict result_info per_page]
		}
		
		
		return $results
	}

	method info {item} {

		switch -exact -- $item {
			zones {
				return [my GetZones]
			}
			default {
				return -code error -errorcode [list tclCloudFlare INFO BADITEM $item] "Unknown info item: '$item'"
			}
		}

		return;
	}

	method zone {id} {
		variable Cache
		variable KeyDict

		if {![dict exists $Cache zone $id]} {
			dict set Cache zone $id [tclCloudFlareZone new [dict get $KeyDict emailAddr] [dict get $KeyDict key] $id]
		}

		return [dict get $Cache zone $id]
	}

}
